#include "lisp.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

void *Vnil;
LISP_DEFVAR(t);
LISP_DEFVAR(module);
LISP_DEFVAR(all_functions);
LISP_DEFVAR(globals);
LISP_DEFVAR(classes);

#ifndef NDEBUG
static size_t _lost_allocs = 0;

static void *lisp_realloc(void *old_ptr, size_t n) {
    if (!old_ptr) {
        ++_lost_allocs;
    }
    void *ptr = realloc(old_ptr, n);
    if (n && !ptr) {
        abort();
    }
    return ptr;
}

// not implemented with realloc to make stack traces more obvious
static void *lisp_malloc(size_t n) {
    void *ptr = malloc(n);
    ++_lost_allocs;
    if (n && !ptr) {
        abort();
    }
    return ptr;
}

static void *lisp_calloc(size_t nmemb, size_t size) {
    ++_lost_allocs;
    void *ptr = calloc(nmemb, size);
    if (size && nmemb && !ptr) {
        abort();
    }
    return ptr;
}

static void lisp_free(void *ptr) {
    --_lost_allocs;
    free(ptr);
}

#define malloc lisp_malloc
#define calloc lisp_calloc
#define free lisp_free
#define realloc lisp_realloc
#endif

LispObject *native_type_symbols[LISP_N_NATIVE_TYPES];
LispSymbol *system_module;
LispInstance *current_error;
StackFrame *call_stack;
size_t call_stack_size;

// Error classes
LispClass *error_class;
LispClass *type_error_class;
LispClass *argument_error_class;
LispClass *function_not_found_error_class;
LispClass *class_slot_error_class;

#define _NATIVE_TYPE_SYMBOL(NAME, name) \
    native_type_symbols[LISP_TYPE_ ## NAME] = ref_lisp_object( \
    INTERN_LITERAL(# name))
static void fill_native_type_symbol_table() {
    _NATIVE_TYPE_SYMBOL(NIL, nil);
    _NATIVE_TYPE_SYMBOL(SYMBOL, symbol);
    _NATIVE_TYPE_SYMBOL(CONS, cons);
    _NATIVE_TYPE_SYMBOL(STRING, string);
    _NATIVE_TYPE_SYMBOL(ARRAY, array);
    _NATIVE_TYPE_SYMBOL(INT, int);
    _NATIVE_TYPE_SYMBOL(FLOAT, float);
    _NATIVE_TYPE_SYMBOL(FUNCTION, function);
    _NATIVE_TYPE_SYMBOL(CLASS, class);
    _NATIVE_TYPE_SYMBOL(HASH_TABLE, hash_table);
    _NATIVE_TYPE_SYMBOL(INSTANCE, instance);
    _NATIVE_TYPE_SYMBOL(POINTER, pointer);
}
#undef _NATIVE_TYPE_SYMBOL

void *ref_lisp_object(void *obj) {
    if (obj && !AS_OBJECT(obj)->persist) {
        ++AS_OBJECT(obj)->ref_count;
    }
    return obj;
}

void *unref_lisp_object(void *obj) {
    if (obj && !AS_OBJECT(obj)->persist && AS_OBJECT(obj)->ref_count <= 1) {
        switch (AS_OBJECT(obj)->type) {
        case LISP_TYPE_SYMBOL:
            unref_lisp_object(((LispSymbol *) obj)->name);
            break;
        case LISP_TYPE_CONS:
            unref_lisp_object(((LispCons *) obj)->car);
            unref_lisp_object(((LispCons *) obj)->cdr);
            break;
        case LISP_TYPE_STRING:
            if (!((LispString *) obj)->skip_free) {
                free(((LispString *) obj)->text);
            }
            break;
        case LISP_TYPE_ARRAY:
            for (size_t i = 0; i < ((LispArray *) obj)->length; ++i) {
                unref_lisp_object(((LispArray *) obj)->data[i]);
            }
            free(((LispArray *) obj)->data);
            break;
        case LISP_TYPE_FUNCTION:
            if (!((LispFunction *) obj)->nativep) {
                unref_lisp_object(((LispFunction *) obj)->lisp.form);
                unref_lisp_object(((LispFunction *) obj)->lisp.pargs);
                unref_lisp_object(((LispFunction *) obj)->lisp.oargs);
                unref_lisp_object(((LispFunction *) obj)->lisp.rarg);
            }
            unref_lisp_object(((LispFunction *) obj)->doc);
            break;
        case LISP_TYPE_HASH_TABLE:
            for (size_t i = 0; i < ((LispHashTable *) obj)->size; ++i) {
                struct LispHashTableBucket *p = ((LispHashTable *) obj)->data[i];
                while (p) {
                    unref_lisp_object(p->key);
                    unref_lisp_object(p->value);
                    struct LispHashTableBucket *next = p->next;
                    free(p);
                    p = next;
                }
            }
            free(((LispHashTable *) obj)->data);
            unref_lisp_object(((LispHashTable *) obj)->hash_func);
            unref_lisp_object(((LispHashTable *) obj)->eq_func);
            break;
        case LISP_TYPE_CLASS:
            unref_lisp_object(((LispClass *) obj)->name);
            unref_lisp_object(((LispClass *) obj)->module);
            unref_lisp_object(((LispClass *) obj)->slots);
            unref_lisp_object(((LispClass *) obj)->constructor);
            unref_lisp_object(((LispClass *) obj)->methods);
            unref_lisp_object(((LispClass *) obj)->superclass);
            break;
        case LISP_TYPE_INSTANCE:
            unref_lisp_object(((LispInstance *) obj)->class);
            unref_lisp_object(((LispInstance *) obj)->slots);
            break;
        case LISP_TYPE_POINTER:
            if (((LispPointer *) obj)->free_func) {
                ((LispPointer *) obj)->free_func(((LispPointer *) obj)->data);
            }
            break;
        case LISP_TYPE_NIL:
        case LISP_TYPE_INT:
        case LISP_TYPE_FLOAT:
            // these don't need any extra steps
            break;
        case LISP_N_NATIVE_TYPES:
            abort();
        }
        free(obj);
        return Vnil;
    } else {
        --AS_OBJECT(obj)->ref_count;
        return obj;
    }
}

void *lisp_check_type(LispType type, void *obj) {
    if (TYPE_OF(obj) != type) {
        ERROR(type_error_class,
              "expected",
              ((LispSymbol *) native_type_symbols[type])->name,
              "got", TYPE_STR(obj));
    }
    return obj;
}

LispObject *call_native(LispObject *(*func)(size_t argc, LispObject **argv),
                        size_t argc, ...) {
    LispObject *arr[argc];
    va_list args;
    va_start(args, argc);
    for (size_t i = 0; i < argc; ++i) {
        arr[i] = va_arg(args, LispObject *);
    }
    va_end(args);
    LispObject *retval = func(argc, arr);
    return retval;
}

LispString *make_lisp_string(const char *text, size_t length, bool skip_free) {
    LispString *obj = malloc(sizeof(LispString));
    obj->parent.type = LISP_TYPE_STRING;
    obj->parent.ref_count = 0;
    obj->parent.persist = false;
    obj->length = length;
    if (skip_free) {
        obj->text = (char *) text;
    } else {
        obj->text = malloc(length + 1);
        memcpy(obj->text, text, length);
        obj->text[length] = '\0';
    }
    obj->skip_free = skip_free;
    return obj;
}

LispInt *make_lisp_int(int64_t value) {
    LispInt *obj = malloc(sizeof(LispInt));
    obj->parent.type = LISP_TYPE_INT;
    obj->parent.ref_count = 0;
    obj->parent.persist = false;
    obj->value = value;
    return obj;
}

LispFloat *make_lisp_float(double value) {
    LispFloat *obj = malloc(sizeof(LispFloat));
    obj->parent.type = LISP_TYPE_INT;
    obj->parent.ref_count = 0;
    obj->parent.persist = false;
    obj->value = value;
    return obj;
}

LispSymbol *make_lisp_symbol(const char *text, size_t length, bool skip_free) {
    LispSymbol *obj = malloc(sizeof(LispSymbol));
    obj->parent.type = LISP_TYPE_SYMBOL;
    obj->parent.ref_count = 0;
    obj->parent.persist = false;
    obj->name = ref_lisp_object(make_lisp_string(text, length, skip_free));
    return obj;
}

LispFunction *make_native_function(bool macrop, size_t nparg, size_t noarg, bool rargp,
                                   LispObject *(*action)(size_t argc,
                                                         LispObject **argv)) {
    LispFunction *obj = malloc(sizeof(LispFunction));
    obj->parent.type = LISP_TYPE_FUNCTION;
    obj->parent.ref_count = 0;
    obj->parent.persist = false;
    obj->doc = Vnil;
    obj->nativep = true;
    obj->macrop = macrop;
    obj->native.action = action;
    obj->native.nparg = nparg;
    obj->native.noarg = noarg;
    obj->native.rargp = rargp;
    return obj;
}

void register_in_module(LispHashTable *table, LispSymbol *module,
                        LispSymbol *name, LispObject *thing) {
    LispHashTable *module_table = (void *) call_native(Fgethash, 2, table, module);
    if (NILP(module_table)) {
        module_table = (LispHashTable *) call_native(Fmake_hash_table, 2, Vnil, Vnil);
        call_native(Fputhash, 3, table, module, module_table);
    }
    call_native(Fputhash, 3, module_table, name, thing);
}

LispObject *lookup_in_module(LispHashTable *table, LispSymbol *module,
                             LispSymbol *name, bool include_system) {
    LispHashTable *module_table = (void *) call_native(Fgethash, 2, table, module);
    if (NILP(module_table)) {
        if (include_system) {
            return lookup_in_module(table, system_module, name, false);
        }
        return Vnil;
    }
    LispObject *result = call_native(Fgethash, 2, module_table, name);
    if (NILP(result) && include_system) {
        result = lookup_in_module(table, system_module, name, false);
    }
    return result;
}

#define register_function(module, name, func) \
    register_in_module((void *) Vall_functions, module, name, (void *) func)
#define lookup_function(module, name) \
    ((LispFunction *) lookup_in_module((void *) Vall_functions, module, name, true))
#define register_class(class) \
    register_in_module((void *) Vclasses, (class)->module, (class)->name, (void *) class)

LispClass *make_lisp_class(LispSymbol *name, LispSymbol *module,
                           LispClass *superclass, LispFunction *constructor,
                           LispCons *slots) {
    LispHashTable *slot_table = (void *) call_native(Fmake_hash_table, 2, Vnil, Vnil); 
    size_t super_high_slot = NILP(superclass) ? 0 : superclass->high_slot;
    size_t i = 0;
    DOLIST(slot, slots, {
        call_native(Fputhash, 3, slot_table, slot,
                    make_lisp_int(super_high_slot + i));
        ++i;
    })
    LispClass *obj = malloc(sizeof(LispClass));
    obj->parent.type = LISP_TYPE_CLASS;
    obj->parent.ref_count = 0;
    obj->parent.persist = false;
    obj->name = ref_lisp_object(name);
    obj->module = ref_lisp_object(module);
    obj->constructor = ref_lisp_object(constructor);
    obj->superclass = ref_lisp_object(superclass);
    obj->high_slot = super_high_slot + slot_table->count;
    obj->slots = ref_lisp_object(slot_table);
    obj->methods = ref_lisp_object(call_native(Fmake_hash_table, 2, Vnil, Vnil));
    return obj;
}

// init_data can be NULL (not nil, actually NULL)
LispArray *make_lisp_array(size_t size, LispObject **init_data) {
    LispArray *obj = malloc(sizeof(LispArray));
    obj->parent.type = LISP_TYPE_ARRAY;
    obj->parent.persist = false;
    obj->parent.ref_count = 0;
    obj->length = size;
    obj->data = calloc(size, sizeof(LispObject *));
    if (init_data) {
        for (size_t i = 0; i < size; ++i) {
            obj->data[i] = ref_lisp_object(init_data[i]);
        }
    } else {
        for (size_t i = 0; i < size; ++i) {
            obj->data[i] = Vnil;
        }
    }
    return obj;
}
#define LISP_ARRAY(array) (make_lisp_array(sizeof(array) / sizeof(LispObject *), \
                           array))

static int64_t lookup_class_slot(LispClass *class, LispSymbol *slot) {
    LispObject *slot_index = Vnil;
    while (NILP(slot_index)) {
        if (NILP(class)) {
            return -1;
        }
        slot_index = call_native(Fgethash, 2, class->slots, slot);
        class = class->superclass;
    }
    return ((LispInt *) slot_index)->value;
}

static void call_constructors(LispClass *class, LispInstance *inst) {
    if (!NILP(class)) {
        call_constructors(class->superclass, inst);
        if (!NILP(class->constructor)) {
            call_native(Ffuncall, 2, class->constructor, inst);
        }
    }
}

LispInstance *make_lisp_instance(LispClass *class, LispHashTable *slots) {
    LispInstance *obj = malloc(sizeof(LispInstance));
    obj->parent.type = LISP_TYPE_INSTANCE;
    obj->parent.persist = false;
    obj->parent.ref_count = 0;
    obj->class = ref_lisp_object(class);
    obj->slots = ref_lisp_object(make_lisp_array(class->high_slot, NULL));
    DOHASH(key, value, slots, {
        int64_t index = lookup_class_slot(class, (LispSymbol *) key);
        if (index < 0) {
            // TODO error
            break;
        }
        obj->slots->data[index] = ref_lisp_object(value);
    })
    call_constructors(class, obj);
    return obj;
}

LispPointer *make_lisp_pointer(void *data, void(*free_func)(void *)) {
    LispPointer *obj = malloc(sizeof(LispPointer));
    obj->parent.type = LISP_TYPE_POINTER;
    obj->parent.persist = false;
    obj->parent.ref_count = 0;
    obj->data = data;
    obj->free_func = free_func;
    return obj;
}

bool instanceof(LispInstance *instance, LispClass *class) {
    LispClass *icls = instance->class;
    while (!NILP(icls)) {
        if (!NILP(call_native(Feq, 2, icls, class))) {
            return true;
        }
        icls = icls->superclass;
    }
    return false;
}

size_t list_length(LispCons *list) {
    size_t len = 0;
    DOLIST(elem, list, {
        ++len;
    })
    return len;
}

void push_stack(LispSymbol *block_name, StackFramePlace *place, jmp_buf *jmp,
                LispHashTable *handlers) {
    StackFrame *frame = malloc(sizeof(StackFrame));
    frame->up = call_stack;
    frame->block_name = ref_lisp_object(block_name);
    frame->jmp = jmp;
    frame->handlers = ref_lisp_object(call_native(Fmake_hash_table, 2,
                                                  INTERN_LITERAL("hash-as-zero"),
                                                  INTERN_LITERAL("subclassp")));
    if (!NILP(handlers)) {
        DOHASH(key, val, handlers, {
            call_native(Fputhash, 3, frame->handlers, key, val);
        })
    }
    if (place) {
        frame->has_place = true;
        frame->place = *place;
        ref_lisp_object(frame->place.file);
    } else {
        frame->has_place= false;
    }
    frame->locals = ref_lisp_object(call_native(Fmake_hash_table, 2, Vnil, Vnil));
    frame->extras = EMPTY_LIST;
    frame->cleanup = EMPTY_LIST;
    ++call_stack_size;
    call_stack = frame;
}

void pop_stack() {
    unref_lisp_object(current_error);
    current_error = Vnil;
    StackFrame *frame = call_stack;
    DOLIST(form, frame->cleanup, {
        call_native(Feval, 1, form);
    })
    unref_lisp_object(frame->block_name);
    unref_lisp_object(frame->handlers);
    unref_lisp_object(frame->locals);
    unref_lisp_object(frame->extras);
    unref_lisp_object(frame->cleanup);
    if (frame->has_place) {
        unref_lisp_object(frame->place.file);
    }
    --call_stack_size;
    call_stack = frame->up;
    free(frame);
}

void add_auto_unref(void *obj) {
    call_stack->extras = CONS(obj, call_stack->extras);
    unref_lisp_object(obj); // because cons refs its arguments
}

void dump_stack(LispInstance *error) {
    LispObject *trace_sym = INTERN_LITERAL("-stack-trace");
    LispCons *trace_list = (void *) call_native(Fslot, 2, error, trace_sym);
    DOLIST(trace_obj, trace_list, {
        LispObject *file_sym = INTERN_LITERAL("file");
        LispObject *block_sym = INTERN_LITERAL("block");
        LispObject *line_sym = INTERN_LITERAL("line");
        LispObject *column_sym = INTERN_LITERAL("column");
        LispString *file = AS_STRING(call_native(Fslot, 2, trace_obj, file_sym));
        LispSymbol *block = AS_SYMBOL(call_native(Fslot, 2, trace_obj, block_sym));
        LispInt *line = AS_INT(call_native(Fslot, 2, trace_obj, line_sym));
        LispInt *column = AS_INT(call_native(Fslot, 2, trace_obj, column_sym));
        // TODO use some lisp printing functions
        fprintf(stderr, "  %*s ", (int) block->name->length, block->name->text);
        fprintf(stderr, "(%*s:%zu:%zu)\n", (int) file->length, file->text,
                (size_t) line->value, (size_t) column->value);
        unref_lisp_object(file_sym);
        unref_lisp_object(block_sym);
        unref_lisp_object(line_sym);
        unref_lisp_object(column_sym);
    })
    unref_lisp_object(trace_sym);
}

void throw_error(LispInstance *error) {
    if (!instanceof(error, error_class)) {
        TYPE_ERROR("error (or a subclass thereof)", TYPE_STR(error));
    }
    StackFrame *current_frame = call_stack;
    while (current_frame) {
        LispInt *handler = (void *) call_native(Fgethash, 2,
                                                current_frame->handlers,
                                                error->class);
        int64_t err_index = NILP(handler) ? 0 : handler->value;
        jmp_buf *jmp = call_stack->jmp;
        if (err_index != 0) {
            while (call_stack != current_frame) {
                pop_stack();
            }
            current_error = ref_lisp_object(error);
            longjmp(*jmp, err_index);
        }
        current_frame = current_frame->up;
    }
    // The error has propagated to the top of the call stack, so we exit...
    // TODO use lisp printing functions
    LispSymbol *message_slot = ref_lisp_object(INTERN_LITERAL("message"));
    LispString *message = (void *) call_native(Fslot, 2, error, message_slot);
    if (STRINGP(message)) {
        fprintf(stderr, "Uncaught %*s: %*s\n",
                (int) error->class->name->name->length,
                error->class->name->name->text,
                (int) message->length, message->text);
    } else {
        fprintf(stderr, "Uncaught %*s: nil\n",
                (int) error->class->name->name->length,
                error->class->name->name->text);
    }
    dump_stack(error);
    exit(1);
}

// Easy interface to create objects and classes
LispClass *new_class(LispClass *superclass, LispSymbol *module,
                     const char *name, bool skip_free, LispFunction *constructor,
                     ...) {
    LispSymbol *name_sym = make_lisp_symbol(name, strlen(name), skip_free);
    LispCons *slot_list = EMPTY_LIST;
    va_list args;
    va_start(args, constructor);
    const char *slot_name = va_arg(args, const char *);
    while (slot_name) {
        slot_list = CONS(make_lisp_symbol(slot_name, strlen(slot_name), true),
                         slot_list);
        slot_name = va_arg(args, const char *);   
    }
    va_end(args);
    LispClass *class = make_lisp_class(name_sym, module, superclass,
                                       constructor, slot_list);
    unref_lisp_object(slot_list);
    return class;
}

LispInstance *new_instance(LispClass *class, ...) {
    va_list args;
    va_start(args, class);
    const char *slot_name = va_arg(args, const char *);
    LispHashTable *slots = (void *) call_native(Fmake_hash_table, 2, Vnil, Vnil);
    while (slot_name) {
        size_t len = strlen(slot_name);
        LispSymbol *slot_sym = make_lisp_symbol(slot_name, len, true);
        LispObject *slot_value = va_arg(args, LispObject *);
        call_native(Fputhash, 3, slots, slot_sym, slot_value);
        slot_name = va_arg(args, const char *);
    }
    LispInstance *obj = make_lisp_instance(class, slots);
    unref_lisp_object(slots);
    va_end(args);
    return obj;
}

#define _REGISTER_NATIVE(lisp_name, name, macrop, nparg, noarg, rargp) \
    register_function(system_module, (void *) INTERN_LITERAL(# lisp_name), \
    make_native_function(macrop, nparg, noarg, rargp, F ## name))
static void register_native_functions() {
    _REGISTER_NATIVE(symbolp, symbolp, false, 1, 0, false);
    _REGISTER_NATIVE(stringp, stringp, false, 1, 0, false);
    _REGISTER_NATIVE(consp, consp, false, 1, 0, false);
    _REGISTER_NATIVE(arrayp, arrayp, false, 1, 0, false);
    _REGISTER_NATIVE(intp, intp, false, 1, 0, false);
    _REGISTER_NATIVE(floatp, floatp, false, 1, 0, false);
    _REGISTER_NATIVE(functionp, functionp, false, 1, 0, false);
    _REGISTER_NATIVE(hash-table-p, hash_table_p, false, 1, 0, false);
    _REGISTER_NATIVE(classp, classp, false, 1, 0, false);
    _REGISTER_NATIVE(instancep, instancep, false, 1, 0, false);

    _REGISTER_NATIVE(listp, listp, false, 1, 0, false);
    _REGISTER_NATIVE(atom, atom, false, 1, 0, false);
    _REGISTER_NATIVE(cons, cons, false, 2, 0, false);
    _REGISTER_NATIVE(intern, intern, false, 1, 0, false);
    _REGISTER_NATIVE(not, not, false, 1, 0, false);
    _REGISTER_NATIVE(hash, hash, false, 1, 0, false);
    _REGISTER_NATIVE(make-hash-table, make_hash_table, false, 0, 2, false);
    _REGISTER_NATIVE(float, float, false, 1, 0, false);
    _REGISTER_NATIVE(eq, eq, false, 2, 0, false);
    _REGISTER_NATIVE(gethash, gethash, false, 2, 0, false);
    _REGISTER_NATIVE(puthash, puthash, false, 3, 0, false);
    _REGISTER_NATIVE(remhash, remhash, false, 2, 0, false);
    _REGISTER_NATIVE(symbol-function, symbol_function, false, 1, 0, false);
    _REGISTER_NATIVE(funcall, funcall, false, 1, 0, true);
    _REGISTER_NATIVE(apply, apply, false, 1, 0, true);
    _REGISTER_NATIVE(eval, eval, false, 1, 0, false);
    _REGISTER_NATIVE(superclassp, superclassp, false, 2, 0, false);
    _REGISTER_NATIVE(subclassp, subclassp, false, 2, 0, false);
    _REGISTER_NATIVE(hash-as-zero, hash_as_zero, false, 1, 0, false);

    // These are special forms, their arguments are NOT evaluated
    _REGISTER_NATIVE(quote, or, true, 0, 0, true);
    _REGISTER_NATIVE(and, and, true, 0, 0, true);
    _REGISTER_NATIVE(or, or, true, 0, 0, true);
}
#undef _REGISTER_NATIVE

static LispClass *lisp_stack_frame_class;

static LispObject *error_constructor(size_t argc, LispObject **argv) {
    LispObject *trace_sym = INTERN_LITERAL("-stack-trace");
    StackFrame *frame = call_stack;
    LispCons *trace_list = EMPTY_LIST;
    while (frame) {
        if (frame->block_name && frame->has_place) {
            LispString *file;
            if (frame->place.is_native) {
                file = STRING_FROM_LITERAL("<native function>");
            } else {
                file = frame->place.file;
            }
            trace_list = CONS(new_instance(lisp_stack_frame_class,
                                           "file", file,
                                           "block", frame->block_name,
                                           "line", make_lisp_int(frame->place.line),
                                           "column", make_lisp_int(frame->place.col),
                                           NULL),
                              trace_list);
        }
        frame = frame->up;
    }
    call_native(Fsetslot, 3, argv[0], trace_sym, trace_list);
    unref_lisp_object(trace_sym);
    return Vnil;
}

static LispObject *type_error_constructor(size_t argc, LispObject **argv) {
    LispInstance *inst = AS_INSTANCE(argv[0]);
    LispObject *message_sym = INTERN_LITERAL("message");
    LispObject *expected_sym = INTERN_LITERAL("expected");
    LispObject *got_sym = INTERN_LITERAL("got");
    LispString *expected = AS_STRING(call_native(Fslot, 2, inst, expected_sym));
    LispString *got = AS_STRING(call_native(Fslot, 2, inst, got_sym));
    // TODO use format
    const char format[] = "expected %s; got %s";
    char msg[sizeof(format) - 4 + expected->length + got->length];
    sprintf(msg, format, expected->text, got->text);
    LispString *msg_obj = make_lisp_string(msg, sizeof(msg) - 1, false);
    call_native(Fsetslot, 3, inst, message_sym, msg_obj);
    unref_lisp_object(message_sym);
    unref_lisp_object(expected_sym);
    unref_lisp_object(got_sym);
    return Vnil;
}

static LispObject *argument_error_constructor(size_t argc, LispObject **argv) {
    LispInstance *inst = AS_INSTANCE(argv[0]);
    LispObject *message_sym = INTERN_LITERAL("message");
    LispObject *expected_sym = INTERN_LITERAL("expected");
    LispObject *got_sym = INTERN_LITERAL("got");
    LispInt *got = AS_INT(call_native(Fslot, 2, inst, got_sym));
    LispObject *expected = call_native(Fslot, 2, inst, expected_sym);
    // TODO use format
    switch (TYPE_OF(expected)) {
    case LISP_TYPE_STRING:
        {
            const char format[] = "expected %s arguments; got %zu";
            char msg[sizeof(format) - 5 + 128 + ((LispString *) expected)->length];
            sprintf(msg, format, ((LispString *) expected)->text, (size_t) got->value);
            LispString *msg_obj = make_lisp_string(msg, sizeof(msg) - 1, false);
            call_native(Fsetslot, 3, inst, message_sym, msg_obj);
        }
        break;
    case LISP_TYPE_NIL:
    default:
        {
            const char format[] = "got %zu arguments";
            char msg[sizeof(format) - 3 + 128];
            sprintf(msg, format, (size_t) got->value);
            LispString *msg_obj = make_lisp_string(msg, sizeof(msg) - 1, false);
            call_native(Fsetslot, 3, inst, message_sym, msg_obj);
        }
        break;
    }
    unref_lisp_object(message_sym);
    unref_lisp_object(expected_sym);
    unref_lisp_object(got_sym);
    return Vnil;
}

static LispObject *function_not_found_error_constructor(size_t argc,
                                                        LispObject **argv) {
    LispInstance *inst = AS_INSTANCE(argv[0]);
    LispObject *message_sym = INTERN_LITERAL("message");
    LispObject *name_sym = INTERN_LITERAL("name");
    LispString *name = AS_STRING(call_native(Fslot, 2, inst, name_sym));
    // TODO use format
    const char format[] = "unknown function: %s";
    char msg[sizeof(format) - 2 + name->length];
    sprintf(msg, format, name->text);
    LispString *msg_obj = make_lisp_string(msg, sizeof(msg) - 1, false);
    call_native(Fsetslot, 3, inst, message_sym, msg_obj);
    unref_lisp_object(name_sym);
    unref_lisp_object(message_sym);
    return Vnil;
}

static LispObject *class_slot_error_constructor(size_t argc, LispObject **argv) {
    LispInstance *inst = AS_INSTANCE(argv[0]);
    LispObject *message_sym = INTERN_LITERAL("message");
    LispObject *class_sym = INTERN_LITERAL("class");
    LispObject *name_sym = INTERN_LITERAL("name");
    LispClass *class = AS_CLASS(call_native(Fslot, 2, inst, class_sym));
    LispString *name = AS_STRING(call_native(Fslot, 2, inst, name_sym));
    // TODO use format
    const char format[] = "class %s has no slot %s";
    char msg[sizeof(format) - 4 + name->length + class->name->name->length];
    sprintf(msg, format, name->text, class->name->name->text);
    LispString *msg_obj = make_lisp_string(msg, sizeof(msg) - 1, false);
    call_native(Fsetslot, 3, inst, message_sym, msg_obj);
    unref_lisp_object(name_sym);
    unref_lisp_object(class_sym);
    unref_lisp_object(message_sym);
    return Vnil;
}

#define _REGISTER_ERROR(cvar, name, ...) \
    cvar ## _class = new_class(error_class, system_module, # name, true, \
                     make_native_function(false, 1, 0, \
                                          false, cvar ## _constructor), \
                     __VA_ARGS__, NULL); \
    register_class(cvar ## _class);

static void register_error_classes() {
    lisp_stack_frame_class = new_class(Vnil, system_module,
                                       "stack-frame", true, Vnil,
                                       "file", "block", "line", "column", NULL);
    register_class(lisp_stack_frame_class);
    
    error_class = new_class(Vnil, system_module,
                            "error", true,
                            make_native_function(false, 1, 0,
                                                 false, error_constructor),
                            "message", "-stack-trace", NULL);
    register_class(error_class);

    _REGISTER_ERROR(type_error, type-error, "expected", "got");
    _REGISTER_ERROR(argument_error, argument-error, "expected", "got");
    _REGISTER_ERROR(function_not_found_error, function-not-found-error, "name");
    _REGISTER_ERROR(class_slot_error, class-slot-error, "class", "name");
}
#undef _REGISTER_ERROR

void init_lisp() {
    // define nil
    Vnil = malloc(sizeof(LispObject));
    ((LispObject *) Vnil)->type = LISP_TYPE_NIL;
    ((LispObject *) Vnil)->persist = true;

    // define t
    Vt = ref_lisp_object(INTERN_LITERAL("t"));

    // The stack...
    call_stack = NULL;
    call_stack_size = 0;

    // Modules
    system_module = ref_lisp_object(INTERN_LITERAL("sl"));
    Vmodule = ref_lisp_object(INTERN_LITERAL("sl-user"));

    // Symbols for native types
    fill_native_type_symbol_table();

    // Important tables
    Vall_functions = call_native(Fmake_hash_table, 2, Vnil, Vnil);
    Vglobals = call_native(Fmake_hash_table, 2, Vnil, Vnil);
    Vclasses = call_native(Fmake_hash_table, 2, Vnil, Vnil);

    // Errors
    current_error = Vnil;
    register_error_classes();

    // Native functions
    register_native_functions();
}

void deinit_lisp() {
    while (call_stack) {
        pop_stack();
    }

    unref_lisp_object(Vclasses);
    unref_lisp_object(Vglobals);
    unref_lisp_object(Vall_functions);
    unref_lisp_object(Vmodule);

    unref_lisp_object(system_module);

    for (size_t i = 0; i < LISP_N_NATIVE_TYPES; ++i) {
        unref_lisp_object(native_type_symbols[i]);
    }

    Vt->persist = false;
    Vt->ref_count = 0;
    ((LispSymbol *) Vt)->name->parent.persist = false;
    ((LispSymbol *) Vt)->name->parent.ref_count = 0;
    unref_lisp_object(Vt);

    free(Vnil);

#ifndef NDEBUG
    if (_lost_allocs) {
        fprintf(stderr, "WARNING: lost %zu allocations!!\n", _lost_allocs);
    }
#endif
}

#define _DEFINE_TYPE_PRED(NAME, name) \
LISP_DEFUN(name ## p) { \
    return LISP_BOOL(NAME ## P (argv[0])); \
}

_DEFINE_TYPE_PRED(SYMBOL, symbol)
_DEFINE_TYPE_PRED(STRING, string)
_DEFINE_TYPE_PRED(CONS, cons)
_DEFINE_TYPE_PRED(ARRAY, array)
_DEFINE_TYPE_PRED(INT, int)
_DEFINE_TYPE_PRED(FLOAT, float)
_DEFINE_TYPE_PRED(FUNCTION, function)
_DEFINE_TYPE_PRED(LIST, list)
_DEFINE_TYPE_PRED(HASH_TABLE_, hash_table_)
_DEFINE_TYPE_PRED(CLASS, class)
_DEFINE_TYPE_PRED(INSTANCE, instance)
_DEFINE_TYPE_PRED(POINTER, pointer)

#undef _DEFINE_TYPE_PRED

LISP_DEFUN(atom) {
    return LISP_BOOL(ATOM(argv[0]));
}

LISP_DEFUN(native) {
    return LISP_BOOL(NATIVE(argv[0]));
}

LISP_DEFUN(type_of) {
    if (NATIVE(argv[0])) {
        return native_type_symbols[argv[0]->type];
    } else {
        return AS_OBJECT(((LispInstance *) argv[0])->class->name);
    }
}

LISP_DEFUN(cons) {
    LispCons *obj = malloc(sizeof(LispCons));
    obj->parent.type = LISP_TYPE_CONS;
    obj->parent.persist = false;
    obj->parent.ref_count = 0;
    obj->car = ref_lisp_object(argv[0]);
    obj->cdr = ref_lisp_object(argv[1]);
    return AS_OBJECT(obj);
}

LISP_DEFUN(list) {
    LispObject *head = Vnil;
    for (size_t i = argc; i >= 1; --i) {
        head = call_native(Fcons, 2, argv[i - 1], head);
    }
    return head;
}

LISP_DEFUN(length) {
    switch (TYPE_OF(argv[0])) {
    case LISP_TYPE_ARRAY:
        return AS_OBJECT(make_lisp_int(((LispArray *) argv[0])->length));
    case LISP_TYPE_STRING:
        return AS_OBJECT(make_lisp_int(((LispString *) argv[0])->length));
    case LISP_TYPE_CONS:
    case LISP_TYPE_NIL:
        return AS_OBJECT(make_lisp_int(list_length((LispCons *) argv[0])));
    case LISP_TYPE_HASH_TABLE:
        return AS_OBJECT(make_lisp_int(((LispHashTable *) argv[0])->count));
    default:
        TYPE_ERROR("array, string, cons, nil, or hash table", TYPE_STR(argv[0]));
        return Vnil;
    }
}

LISP_DEFUN(intern) {
    LispString *name = AS_STRING(argv[0]);
    LispSymbol *obj = malloc(sizeof(LispSymbol));
    obj->parent.type = LISP_TYPE_SYMBOL;
    obj->parent.persist = false;
    obj->parent.ref_count = 0;
    obj->name = ref_lisp_object(name);
    return AS_OBJECT(obj);
}

LISP_DEFUN(not) {
    return LISP_BOOL(NILP(argv[0]));
}

LISP_DEFUN(hash) {
    int64_t h = 37;
    switch (TYPE_OF(argv[0])) {
    case LISP_TYPE_CLASS:
        {
            LispClass *cls = (void *) argv[0];
            LispInt *one = ref_lisp_object(call_native(Fhash, 1, cls->module));
            LispInt *two = ref_lisp_object(call_native(Fhash, 1, cls->name));
            LispInt *retval = make_lisp_int(one->value + two->value);
            unref_lisp_object(one);
            unref_lisp_object(two);
            return AS_OBJECT(retval);
        }
    case LISP_TYPE_SYMBOL:
        argv[0] = AS_OBJECT(((LispSymbol *) argv[0])->name);
    case LISP_TYPE_STRING:
        for (size_t i = 0; i < ((LispString *) argv[0])->length; ++i) {
            h = (h * 54059) ^ (((LispString *) argv[0])->text[i] * 76963);
        }
        return AS_OBJECT(make_lisp_int(h % 86969));
    default:
        return AS_OBJECT(make_lisp_int(0));
    }
}

LISP_DEFUN(make_hash_table) {
    if (!NILP(argv[0]) && !SYMBOLP(argv[0]) && !FUNCTIONP(argv[0])) {
        TYPE_ERROR("function, symbol, or nil", TYPE_STR(argv[0]));
    }
    if (!NILP(argv[1]) && !SYMBOLP(argv[1]) && !FUNCTIONP(argv[1])) {
        TYPE_ERROR("function, symbol, or nil", TYPE_STR(argv[1]));
    }
    LispHashTable *obj = malloc(sizeof(LispHashTable));
    obj->parent.type = LISP_TYPE_HASH_TABLE;
    obj->parent.ref_count = 0;
    obj->parent.persist = false;
    obj->size = LISP_HASH_TABLE_INITIAL_SIZE;
    obj->count = 0;
    obj->data = calloc(obj->size, sizeof(struct LispHashTableBucket));
    obj->hash_func = ref_lisp_object(argv[0]);
    obj->eq_func = ref_lisp_object(argv[1]);
    return AS_OBJECT(obj);
}

LISP_DEFUN(float) {
    switch (TYPE_OF(argv[0])) {
    case LISP_TYPE_INT:
        return AS_OBJECT(make_lisp_float(((LispInt *) argv[0])->value));
    case LISP_TYPE_FLOAT:
        return argv[0];
    default:
        TYPE_ERROR("float or int", TYPE_OF(argv[0]));
        return Vnil;
    }
}

LISP_DEFUN(eq) {
    if ((INTP(argv[0]) || FLOATP(argv[0])) &&
        (INTP(argv[1]) || FLOATP(argv[1]))) {
        LispFloat *f1 = ref_lisp_object(call_native(Ffloat, 1, argv[0]));
        LispFloat *f2 = ref_lisp_object(call_native(Ffloat, 1, argv[1]));
        bool equal = f1->value == f2->value;
        unref_lisp_object(f1);
        unref_lisp_object(f2);
        return LISP_BOOL(equal);
    } else if (TYPE_OF(argv[0]) == TYPE_OF(argv[1])) {
        switch (TYPE_OF(argv[0])) {
        case LISP_TYPE_NIL:
            return Vt;
        case LISP_TYPE_SYMBOL:
            return LISP_BOOL(strncmp(((LispSymbol *) argv[0])->name->text,
                                     ((LispSymbol *) argv[1])->name->text,
                                     ((LispSymbol *) argv[1])->name->length)
                             == 0);
        case LISP_TYPE_CLASS:
            {
                LispObject *one = call_native(Feq, 2, ((LispClass *) argv[0])->name,
                                              ((LispClass *) argv[1])->name);
                LispObject *two = call_native(Feq, 2, ((LispClass *) argv[0])->module,
                                              ((LispClass *) argv[1])->module);
                return LISP_BOOL(!NILP(one) && !NILP(two));
            }
        case LISP_TYPE_FUNCTION:
            if (((LispFunction *) argv[0])->nativep) {
                return LISP_BOOL(((LispFunction *) argv[0])->native.action ==
                                 ((LispFunction *) argv[1])->native.action);
            }
            // fall-through
        default:
            return LISP_BOOL(argv[0] == argv[1]);
        }
    }
    return Vnil;
}

static uint64_t hash_obj_for_table(LispHashTable *table, LispObject *obj) {
    LispInt *hash_obj;
    if (NILP(table->hash_func)) {
        hash_obj = (void *) call_native(Fhash, 1, obj);
    } else {
        hash_obj = (void *) call_native(Ffuncall, 2, table->hash_func, obj);
    }
    uint64_t hash = hash_obj->value;
    unref_lisp_object(hash_obj);
    return hash;
}

static struct LispHashTableBucket *find_bucket_in_table(LispHashTable *table,
                                                         uint64_t hash,
                                                         LispObject *obj) {
    struct LispHashTableBucket *bucket = table->data[hash % table->size];
    while (bucket) {
        bool equal;
        if (NILP(table->eq_func)) {
            equal = !NILP(call_native(Feq, 2, obj, bucket->key));
        } else {
            equal = !NILP(call_native(Ffuncall, 3, table->eq_func, obj, bucket->key));
        }
        if (equal) {
            break;
        }
        bucket = bucket->next;
    }
    return bucket;
}

LISP_DEFUN(gethash) {
    LispHashTable *table = AS_HASH_TABLE(argv[0]);
    uint64_t hash = hash_obj_for_table(table, argv[1]);
    struct LispHashTableBucket *bucket = find_bucket_in_table(table, hash, argv[1]);
    return bucket ? bucket->value : Vnil;
}

static void size_up_hash_table(LispHashTable *table) {
    size_t old_size = table->size;
    struct LispHashTableBucket **old_buckets = table->data;
    table->size *= 2;
    table->data = calloc(table->size, sizeof(struct LispHashTableBucket *));
    for (size_t i = 0; i < old_size; ++i) {
        struct LispHashTableBucket *bucket = old_buckets[i];
        while (bucket) {
            struct LispHashTableBucket *next = bucket->next;
            bucket->next = table->data[next->hash % table->size];
            table->data[next->hash % table->size] = bucket;
            bucket = next;
        }
    }
    free(old_buckets);
}

LISP_DEFUN(puthash) {
    LispHashTable *table = AS_HASH_TABLE(argv[0]);
    if (table->count == table->size) {
        size_up_hash_table(table);
    }

    uint64_t hash = hash_obj_for_table(table, argv[1]);
    struct LispHashTableBucket *bucket = find_bucket_in_table(table, hash, argv[1]);
    if (bucket) {
        LispObject *old_value = unref_lisp_object(bucket->value);
        bucket->value = argv[1];
        return old_value;
    }
    struct LispHashTableBucket *nb = malloc(sizeof(struct LispHashTableBucket));
    if (table->data[hash % table->size]) {
        table->data[hash % table->size]->prev = nb;
    }
    nb->next = table->data[hash % table->size];
    nb->prev = NULL;
    nb->hash = hash;
    nb->key = ref_lisp_object(argv[1]);
    nb->value = ref_lisp_object(argv[2]);
    table->data[hash % table->size] = nb;
    ++table->count;
    return argv[2];
}

LISP_DEFUN(remhash) {
    LispHashTable *table = AS_HASH_TABLE(argv[0]);
    uint64_t hash = hash_obj_for_table(table, argv[1]);
    struct LispHashTableBucket *bucket = find_bucket_in_table(table, hash, argv[1]);
    if (bucket) {
        LispObject *old_value = bucket->value;
        bucket->prev->next = bucket->next;
        bucket->next->prev = bucket->prev;
        unref_lisp_object(bucket->key);
        free(bucket);
        return unref_lisp_object(old_value);
    }
    return Vnil;
}

LISP_DEFUN(symbol_function) {
    LispSymbol *sym = AS_SYMBOL(argv[0]);
    return AS_OBJECT(lookup_function((LispSymbol *) Vmodule, sym));
}

static bool check_function_arguments(LispFunction *func, size_t argc) {
    size_t nparg, noarg;
    bool rargp;
    if (func->nativep) {
        nparg = func->native.nparg;
        noarg = func->native.noarg;
        rargp = func->native.rargp;
    } else {
        nparg = func->lisp.pargs->length;
        noarg = func->lisp.oargs->length;
        rargp = !NILP(func->lisp.rarg);
    }
    return argc >= nparg && (rargp || argc <= nparg + noarg);
}

LISP_DEFUN(funcall) {
    LispFunction *func;
    if (FUNCTIONP(argv[0])) {
        func = (void *) argv[0];
    } else if (SYMBOLP(argv[0])) {
        func = lookup_function((LispSymbol *) Vmodule, (LispSymbol *) argv[0]);
        if (NILP(func)) {
            ERROR(function_not_found_error_class,
                  "name", ((LispSymbol *) argv[0])->name);
        }
    } else {
        TYPE_ERROR("function or symbol", TYPE_STR(argv[0]));
    }
    ref_lisp_object(func);
    push_stack(Vnil, NULL, NULL, Vnil);
    add_auto_unref(func);
    if (!check_function_arguments(func, argc - 1)) {
        ERROR(argument_error_class, "got", make_lisp_int(argc - 1));
    }
    LispObject *retval;
    if (func->nativep) {
        retval = func->native.action(argc - 1, &argv[1]);
    } else {
        // TODO call the lisp function
    }
    pop_stack();
    return retval;
}

LISP_DEFUN(apply) {
    if (NILP(argv[argc - 1]) || !LISTP(argv[argc - 1])) {
        return Ffuncall(argc, argv);
    }
    LispArray *rest_args = (void *) call_native(Flist_to_array, 1, argv[argc - 1]);
    size_t real_argc = argc - 1 + rest_args->length;
    LispObject *real_argv[real_argc];
    for (size_t i = 0; i < argc - 1; ++i) {
        real_argv[i] = argv[i];
    }
    for (size_t i = 0; i < rest_args->length; ++i) {
        real_argv[argc - 1 + i] = rest_args->data[i];
    }
    unref_lisp_object(rest_args);
    return Ffuncall(real_argc, real_argv);
}

LISP_DEFUN(eval) {
    if (ATOM(argv[0])) {
        return argv[0];
    }
    // TODO implement
    return Vnil;
}

LISP_DEFUN(make_instance) {
    LispClass *class;
    if (CLASSP(argv[0])) {
        class = (void *) argv[0];
    } else if (SYMBOLP(argv[0])) {
        class = (void *) lookup_in_module((LispHashTable *) Vclasses,
                                          (LispSymbol *) Vmodule,
                                          (LispSymbol *) argv[0],
                                          true);
    } else {
        TYPE_ERROR("class or symbol", TYPE_STR(argv[0]));
    }
    if (argc % 2 != 1) {
        ERROR(argument_error_class,
              "expected",
              STRING_FROM_LITERAL("even number of keys for make-instance"),
              "got", make_lisp_int(argc - 1));
    }
    LispHashTable *slots = (void *) call_native(Fmake_hash_table, 2, Vnil, Vnil);
    ref_lisp_object(slots);
    for (size_t i = 1; i < argc - 1; i += 2) {
        if (!SYMBOLP(argv[i])) {
            // TODO error
            continue;
        }
        call_native(Fputhash, 3, slots, argv[i], argv[i + 1]);
    }
    LispInstance *obj = make_lisp_instance(class, slots);
    unref_lisp_object(slots);
    return AS_OBJECT(obj);
}

LISP_DEFUN(slot) {
    LispInstance *inst = AS_INSTANCE(argv[0]);
    LispSymbol *name = AS_SYMBOL(argv[1]);
    int64_t index = lookup_class_slot(inst->class, name);
    if (index < 0) {
        ERROR(class_slot_error_class,
              "class", inst->class,
              "name", name->name);
    }
    return inst->slots->data[index];
}

LISP_DEFUN(setslot) {
    LispInstance *inst = AS_INSTANCE(argv[0]);
    LispSymbol *name = AS_SYMBOL(argv[1]);
    int64_t index = lookup_class_slot(inst->class, name);
    if (index < 0) {
        ERROR(class_slot_error_class,
              "class", inst->class,
              "name", name->name);
    }
    LispObject *old_value = unref_lisp_object(inst->slots->data[index]);
    inst->slots->data[index] = ref_lisp_object(argv[2]);
    return old_value;
}

LISP_DEFUN(array_to_list) {
    LispArray *arr = AS_ARRAY(argv[0]);
    LispCons *out = EMPTY_LIST;
    DOARRAY(i, elem, arr, {
        out = CONS(elem, out);
    })
    return AS_OBJECT(out);
}

LISP_DEFUN(list_to_array) {
    if (NILP(argv[0])) {
        return AS_OBJECT(make_lisp_array(0, NULL));
    }
    LispCons *list = AS_CONS(argv[0]);
    size_t list_len = list_length(list);
    LispArray *arr = make_lisp_array(list_len, NULL);
    size_t i = 0;
    DOLIST(elem, (LispCons *) argv[0], {
        arr->data[i++] = ref_lisp_object(elem);
    })
    return AS_OBJECT(arr);
}

// (defun superclassp parent class)
LISP_DEFUN(superclassp) {
    LispClass *parent = AS_CLASS(argv[0]);
    LispClass *child = AS_CLASS(argv[1]);
    while (!NILP(child)) {
        if (!NILP(call_native(Feq, 2, parent, child))) {
            return Vt;
        }
        child = child->superclass;
    }
    return Vnil;
}

// (defun superclassp child class)
LISP_DEFUN(subclassp) {
    return call_native(Fsuperclassp, 2, argv[1], argv[0]);
}

LISP_DEFUN(hash_as_zero) {
    return AS_OBJECT(make_lisp_int(0));
}

LISP_DEFUN(quote) {
    return argv[0];
}

LISP_DEFUN(and) {
    for (size_t i = 0; i < argc; ++i) {
        LispObject *retval = ref_lisp_object(Feval(1, &argv[i]));
        bool is_nil = NILP(retval);
        unref_lisp_object(retval);
        if (is_nil) {
            return Vnil;
        }
    }
    return Vt;
}

LISP_DEFUN(or) {
    for (size_t i = 0; i < argc; ++i) {
        LispObject *retval = Feval(1, &argv[i]);
        if (!NILP(retval)) {
            return retval;
        }
    }
    return Vnil;
}
