#include <stddef.h>
#include <stdbool.h>
#include <stdint.h>
#include <setjmp.h>

typedef enum {
    LISP_TYPE_NIL = 0, // only one value, nil (unit type)
    LISP_TYPE_SYMBOL,
    LISP_TYPE_CONS,
    LISP_TYPE_STRING,
    LISP_TYPE_ARRAY,
    LISP_TYPE_INT,
    LISP_TYPE_FLOAT,
    LISP_TYPE_FUNCTION,
    LISP_TYPE_CLASS,
    LISP_TYPE_HASH_TABLE,
    LISP_TYPE_INSTANCE,
    LISP_TYPE_POINTER,
    LISP_N_NATIVE_TYPES,
} LispType;

typedef struct {
    LispType type;
    bool persist;
    size_t ref_count;
} LispObject;

typedef struct {
    LispObject parent;
    bool skip_free;
    char *text;
    size_t length;
} LispString;

typedef struct {
    LispObject parent;
    LispString *name;
} LispSymbol;

typedef struct {
    LispObject parent;
    LispObject *car;
    LispObject *cdr;
} LispCons;
#define CONS(car, cdr) ((void *) call_native(Fcons, 2, (car), (cdr)))
#define EMPTY_LIST ((LispCons *) Vnil)

#define DOLIST(cvar, list, body) \
    for (LispCons *__dolist_c = list; CONSP(__dolist_c); __dolist_c = (void *) \
         __dolist_c->cdr) { \
        LispObject * cvar = __dolist_c->car; \
        body \
    }

typedef struct {
    LispObject parent;
    LispObject **data;
    size_t length;
} LispArray;

#define DOARRAY(index, value, arr, body) \
    for (size_t index = 0; index < (arr)->length; ++ index) { \
        LispObject * value = (arr)->data[ index ]; \
        body \
    }

typedef struct {
    LispObject parent;
    int64_t value;
} LispInt;

typedef struct {
    LispObject parent;
    double value;
} LispFloat;



typedef struct {
    LispObject parent;
    LispString *doc;
    bool nativep;
    bool macrop;
    union {
        struct {
            LispObject *(*action)(size_t argc, LispObject **argv);
            size_t nparg;
            size_t noarg;
            bool rargp;
        } native;
        struct {
            LispObject *form;
            LispArray *pargs; // list of position arguments
            LispArray *oargs; // list of optional positional arguments
            LispSymbol *rarg; // name of the rest argument
        } lisp;
    };
} LispFunction;

#define LISP_HASH_TABLE_INITIAL_SIZE 32

struct LispHashTableBucket {
    uint64_t hash;
    LispObject *key;
    LispObject *value;
    struct LispHashTableBucket *next;
    struct LispHashTableBucket *prev;
};

typedef struct {
    LispObject parent;
    size_t size;
    size_t count;
    struct LispHashTableBucket **data;
    LispObject *hash_func;
    LispObject *eq_func;
} LispHashTable;

#define DOHASH(kvar, vvar, table, body) \
    for (size_t __dohash_i = 0; __dohash_i < (table)->size; ++__dohash_i) { \
        struct LispHashTableBucket *__dohash_bucket = (table)->data[__dohash_i]; \
        while (__dohash_bucket) { \
            LispObject * kvar = __dohash_bucket->key; \
            LispObject * vvar = __dohash_bucket->value; \
            body \
            __dohash_bucket = __dohash_bucket->next; \
        } \
    }

typedef struct _LispClass {
    LispObject parent;
    struct _LispClass *superclass;
    LispSymbol *name;
    LispSymbol *module;
    LispFunction *constructor;
    LispHashTable *slots; // (name . index)
    size_t high_slot;
    LispHashTable *methods; // (name . function)
} LispClass;

typedef struct {
    LispObject parent;
    LispClass *class;
    LispArray *slots;
} LispInstance;

typedef struct {
    LispObject parent;
    void *data;
    void (*free_func)(void *);
} LispPointer;

extern LispObject *native_type_symbols[LISP_N_NATIVE_TYPES];
extern LispSymbol *system_module;
extern LispInstance *current_error;

// Error classes
extern LispClass *error_class;
extern LispClass *type_error_class;
extern LispClass *argument_error_class;
extern LispClass *function_not_found_error_class;
extern LispClass *class_slot_error_class;

typedef struct {
    LispString *file;
    bool is_native;
    size_t line;
    size_t col;
} StackFramePlace;

typedef struct _StackFrame {
    struct _StackFrame *up;
    LispSymbol *block_name;
    // (error class . int (to return from setjmp)) or nil
    bool has_place;
    StackFramePlace place;
    LispHashTable *handlers; 
    jmp_buf *jmp;
    LispHashTable *locals;
    LispCons *extras; // extra objects to be unrefed
    LispCons *cleanup; // list of cleanup forms
} StackFrame;

extern StackFrame *call_stack;
extern size_t call_stack_size;

#define AS_OBJECT(obj) ((LispObject *) (obj))
#define LISP_BOOL(obj) ((obj) ? Vt : Vnil)
#define TYPE_OF(obj) (AS_OBJECT(obj)->type)
#define COUNT_REFS(obj) (AS_OBJECT(obj)->ref_count)

#define LISP_DEFVAR(name) LispObject * V ## name

extern void *Vnil;
extern LISP_DEFVAR(t); // symbol
extern LISP_DEFVAR(module);
extern LISP_DEFVAR(all_functions); // hash table (symbol . (symbol . function))
extern LISP_DEFVAR(globals); // hash table (symbol . (symbol . any))
extern LISP_DEFVAR(classes); // hash table (symbol . (symbol. class))

void *ref_lisp_object(void *obj);
void *unref_lisp_object(void *obj);
void *lisp_check_type(LispType type, void *obj);
LispObject *call_native(LispObject *(*func)(size_t argc, LispObject **argv),
                        size_t argc, ...);
LispString *make_lisp_string(const char *text, size_t length,
                             bool skip_free);
LispInt *make_lisp_int(int64_t value);
LispFloat *make_lisp_float(double value);
LispSymbol *make_lisp_symbol(const char *text, size_t length, bool skip_free);
LispFunction *make_native_function(bool macrop, size_t nparg, size_t noarg, bool rargp,
                                   LispObject *(*action)(size_t argc, LispObject **argv));
LispObject *lookup_in_module(LispHashTable *table, LispSymbol *module,
                             LispSymbol *name, bool include_system);
void register_in_module(LispHashTable *table, LispSymbol *module,
                        LispSymbol *name, LispObject *thing);
LispArray *make_lisp_array(size_t size, LispObject **init_data);
LispClass *make_lisp_class(LispSymbol *name, LispSymbol *module,
                           LispClass *superclass, LispFunction *constructor,
                           LispCons *slots);
LispInstance *make_lisp_instance(LispClass *class, LispHashTable *slots);
LispPointer *make_lisp_pointer(void *data, void(*free_func)(void *));
void push_stack(LispSymbol *block_name, StackFramePlace *place, jmp_buf *jmp,
                LispHashTable *handlers);
void pop_stack(void);
void add_auto_unref(void *obj);
void dump_stack(LispInstance *error);
void throw_error(LispInstance *error);
LispClass *new_class(LispClass *superclass, LispSymbol *module,
                     const char *name, bool skip_free, LispFunction *constructor,
                     ...);
LispInstance *new_instance(LispClass *class, ...);
#define ERROR(...) throw_error(new_instance(__VA_ARGS__, NULL))
#define TYPE_ERROR(exp, got) (ERROR(type_error_class, "expected", \
                                    STRING_FROM_LITERAL(exp), "got", (got)))
bool instanceof(LispInstance *instance, LispClass *class);
size_t list_length(LispCons *list);

#define INTERN_LITERAL(name) \
    (AS_OBJECT(make_lisp_symbol(name, sizeof(name) - 1, true)))
#define STRING_FROM_LITERAL(text) (make_lisp_string(text, sizeof(text) - 1, true))
void init_lisp(void);
void deinit_lisp(void);

#define LISP_DEFUN(name) LispObject * F ## name (size_t argc, LispObject **argv)

#define NILP(obj) (TYPE_OF(obj) == LISP_TYPE_NIL)

#define SYMBOLP(obj) (TYPE_OF(obj) == LISP_TYPE_SYMBOL)
#define AS_SYMBOL(obj) ((LispSymbol *) lisp_check_type(LISP_TYPE_SYMBOL, obj))
LISP_DEFUN(symbolp);

#define STRINGP(obj) (TYPE_OF(obj) == LISP_TYPE_STRING)
#define AS_STRING(obj) ((LispString *) lisp_check_type(LISP_TYPE_STRING, obj))
LISP_DEFUN(stringp);

#define CONSP(obj) (TYPE_OF(obj) == LISP_TYPE_CONS)
#define AS_CONS(obj) ((LispCons *) lisp_check_type(LISP_TYPE_CONS, obj))
LISP_DEFUN(consp);

#define ARRAYP(obj) (TYPE_OF(obj) == LISP_TYPE_ARRAY)
#define AS_ARRAY(obj) ((LispArray *) lisp_check_type(LISP_TYPE_ARRAY, obj))
LISP_DEFUN(arrayp);

#define INTP(obj) (TYPE_OF(obj) == LISP_TYPE_INT)
#define AS_INT(obj) ((LispInt *) lisp_check_type(LISP_TYPE_INT, obj))
LISP_DEFUN(intp);

#define FLOATP(obj) (TYPE_OF(obj) == LISP_TYPE_FLOAT)
#define AS_FLOAT(obj) ((LispFloat *) lisp_check_type(LISP_TYPE_FLOAT, obj))
LISP_DEFUN(floatp);

#define FUNCTIONP(obj) (TYPE_OF(obj) == LISP_TYPE_FUNCTION)
#define AS_FUNCTION(obj) ((LispFunction *) lisp_check_type(LISP_TYPE_FUNCTION, obj))
LISP_DEFUN(functionp);

#define HASH_TABLE_P(obj) (TYPE_OF(obj) == LISP_TYPE_HASH_TABLE)
#define AS_HASH_TABLE(obj) ((LispHashTable *) lisp_check_type(LISP_TYPE_HASH_TABLE, obj))
LISP_DEFUN(hash_table_p);

#define CLASSP(obj) (TYPE_OF(obj) == LISP_TYPE_CLASS)
#define AS_CLASS(obj) ((LispClass *) lisp_check_type(LISP_TYPE_CLASS, obj))
LISP_DEFUN(classp);

#define INSTANCEP(obj) (TYPE_OF(obj) == LISP_TYPE_INSTANCE)
#define AS_INSTANCE(obj) ((LispInstance *) lisp_check_type(LISP_TYPE_INSTANCE, obj))
LISP_DEFUN(instancep);

#define POINTERP(obj) (TYPE_OF(obj) == LISP_TYPE_POINTER)
#define AS_POINTER(obj) ((LispPointer *) lisp_check_type(LISP_TYPE_POINTER, obj))
LISP_DEFUN(pointerp);

// NOTE: this evaluates obj twice!!
#define LISTP(obj) (NILP(obj) || CONSP(obj))
LISP_DEFUN(listp);

#define ATOM(obj) (!CONSP(obj))
LISP_DEFUN(atom);

#define NATIVE(obj) (!INSTANCEP(obj))
LISP_DEFUN(native);

LISP_DEFUN(type_of);
#define TYPE_STR(obj) (((LispSymbol *) call_native(Ftype_of, 1, (obj)))->name)
LISP_DEFUN(cons);
LISP_DEFUN(list);
LISP_DEFUN(length);
LISP_DEFUN(intern);
LISP_DEFUN(not);
LISP_DEFUN(hash);
LISP_DEFUN(make_hash_table);
LISP_DEFUN(float);
LISP_DEFUN(eq);
LISP_DEFUN(gethash);
LISP_DEFUN(puthash);
LISP_DEFUN(remhash);
LISP_DEFUN(symbol_function);
LISP_DEFUN(funcall);
LISP_DEFUN(apply);
LISP_DEFUN(eval);
LISP_DEFUN(make_instance);
LISP_DEFUN(slot);
LISP_DEFUN(setslot);
LISP_DEFUN(array_to_list);
LISP_DEFUN(list_to_array);
LISP_DEFUN(superclassp);
LISP_DEFUN(subclassp);
LISP_DEFUN(hash_as_zero);

// These are special forms, they their argument are NOT evaluated
LISP_DEFUN(quote);
LISP_DEFUN(and);
LISP_DEFUN(or);
